@@ -1,10 +1,10 @@
-"""f90nml.py
-Parse fortran namelist files into dicts of standard Python data types.
-Contact: Marshall Ward <python@marshallward.org>
----
-Distributed as part of f90nml, Copyright 2014 Marshall Ward
-Licensed under the Apache License, Version 2.0
-http://www.apache.org/licenses/LICENSE-2.0
+"""f90nml
+   ======
+
+   A Fortran 90 namelist parser and generator.
+
+   :copyright: Copyright 2014 Marshall Ward <python@marshallward.org>
+   :license: Apache License, Version 2.0, see LICENSE for details.
 """
 
 from collections import OrderedDict
@@ -12,7 +12,7 @@ import itertools
 import os
 import shlex
 
-__version__ = '0.4'
+__version__ = '0.5'
 
 
 #---
@@ -45,95 +45,44 @@ def read(nml_fname, verbose=False):
         g_vars = NmlDict()
 
         v_name = None
-        v_idx = None
-        v_vals = []
 
         # Current token is either a variable name or finalizer (/, &)
 
         # Populate the namelist group
         while g_name:
 
-            if not v_name:
+            if not t in ('=', '%', '('):
                 t, prior_t = next(tokens), t
 
                 # Skip commas separating objects
                 if t == ',':
                     t, prior_t = next(tokens), t
-                # Parse the indices of the current variable
-            else:
-                if t == '(':
-                    v_name, v_indices, t = parse_f90idx(tokens, t, prior_t)
-
-                    # TODO: Multidimensional support
-                    # TODO: End index support (currently ignored)
-                    i_s = 1 if not v_indices[0][0] else v_indices[0][0]
-                    i_r = 1 if not v_indices[0][2] else v_indices[0][2]
-                    v_idx = itertools.count(i_s, i_r)
 
             # Diagnostic testing
             if verbose:
                 print('  tokens: {} {}'.format(prior_t, t))
 
-            while v_name:
-
-                t, prior_t = next(tokens), t
-
-                if verbose:
-                    print('    vstate: {} {} {}'.format(v_name, v_idx, v_vals))
-                    print('    tokens: {} {}'.format(prior_t, t))
-
-                # Parse the prior token value
-                # TODO: Add '%' to first tuple
-                if (not t in ('(', '=') or prior_t == '=') \
-                        and not (prior_t, t) == (',', '/'):
-                    # Parse the variable string
-                    if prior_t in ('=', ','):
-                        if t in (',', '/', '&'):
-                            next_value = None
-                        else:
-                            continue
-                    else:
-                        next_value, t = parse_f90val(tokens, t, prior_t)
-
-                    if v_idx:
-
-                        v_i = next(v_idx)
-
-                        if v_name in g_vars:
-                            v_vals = g_vars[v_name]
-                            if type(v_vals) != list:
-                                v_vals = [v_vals]
-
-                        try:
-                            # Default Fortran indexing starts at 1
-                            v_vals[v_i - 1] = next_value
-                        except IndexError:
-                            # Expand list to accomodate out-of-range indices
-                            size = len(v_vals)
-                            v_vals.extend(None for i in range(size, v_i))
-                            v_vals[v_i - 1] = next_value
-                    else:
-                        v_vals.append(next_value)
-
-                # Save then deactivate the current variable
-                # TODO: Add '%'
-                if t in ('(', '=', '/', '&'):
-
-                    if len(v_vals) == 0:
-                        v_vals = None
-                    elif len(v_vals) == 1:
-                        v_vals = v_vals[0]
-
-                    g_vars[v_name] = v_vals
-
-                    v_name = None
-                    v_vals = []
-
             # Set the next active variable
-            # TODO: Add '%' to the list
-            if t in ('=', '('):
-                v_name = prior_t
-                v_idx = None
+            if t in ('=', '(', '%'):
+                v_name, v_values, t, prior_t = parse_f90var(tokens, t, prior_t)
+
+                if v_name in g_vars:
+                    v_prior_values = g_vars[v_name]
+                    if type(v_prior_values) != list:
+                        v_prior_values = [v_prior_values]
+
+                    v_values = merge_values(v_prior_values, v_values)
+
+                if len(v_values) == 0:
+                    v_values = None
+                elif len(v_values) == 1:
+                    v_values = v_values[0]
+
+                g_vars[v_name] = v_values
+
+                # Deselect variable
+                v_name = None
+                v_values = []
 
             # Finalise namelist group
             if t in ('/', '&'):
@@ -164,22 +113,122 @@ def write(nml, nml_fname, force=False):
         nml_file.write('&{}\n'.format(grp))
 
         grp_vars = nml[grp]
-        for v_name in grp_vars.keys():
+        for v_name, v_val in grp_vars.items():
 
-            v_val = grp_vars[v_name]
-
-            if type(v_val) == list:
-                v_str = ', '.join([to_f90str(v) for v in v_val])
-            else:
-                v_str = to_f90str(v_val)
-
-            nml_file.write('    {} = {}\n'.format(v_name, v_str))
+            for v_str in var_strings(v_name, v_val):
+                nml_file.write('    {}\n'.format(v_str))
 
         nml_file.write('/\n')
 
     nml_file.close()
 
 
+#---
+def var_strings(v_name, v_values):
+
+    var_strs = []
+
+    if type(v_values) in (dict, NmlDict):
+
+        for f_name, f_vals in v_values.items():
+            var_strs.append('%'.join([v_name, var_strings(f_name, f_vals)[0]]))
+
+    else:
+        if type(v_values) == list:
+            v_vals = ', '.join([to_f90str(v) for v in v_values])
+        else:
+            v_vals = to_f90str(v_values)
+
+        var_strs.append('{} = {}'.format(v_name, v_vals))
+
+    return var_strs
+
+
+#---
+def parse_f90var(tokens, t, prior_t):
+    """Parse a variable and return its name and values."""
+
+    v_name = prior_t
+    v_values = []
+
+    # Parse the indices of the current variable
+    if t == '(':
+        v_indices, t = parse_f90idx(tokens, t, prior_t)
+
+        # TODO: Multidimensional support
+        i_s = 1 if not v_indices[0][0] else v_indices[0][0]
+        i_e = v_indices[0][1]
+        i_r = 1 if not v_indices[0][2] else v_indices[0][2]
+
+        if i_e:
+            v_idx = iter(range(i_s, i_e, i_r))
+        else:
+            v_idx = itertools.count(i_s, i_r)
+    else:
+        v_idx = None
+
+    if t == '%':
+        # Resolve the derived type
+        t, prior_t = next(tokens), t
+        t, prior_t = next(tokens), t
+        v_att, v_att_vals, t, prior_t = parse_f90var(tokens, t, prior_t)
+
+        # TODO: resolve indices
+        next_value = NmlDict()
+
+        if len(v_att_vals) == 0:
+            v_att_vals = None
+        elif len(v_att_vals) == 1:
+            v_att_vals = v_att_vals[0]
+
+        next_value[v_att] = v_att_vals
+
+        append_value(v_values, next_value, v_idx)
+
+    else:
+        # Construct the variable array
+        assert t == '='
+        t, prior_t = next(tokens), t
+
+        # Add variables until next variable trigger (excepting complex tokens)
+        while not t in ('=', '(', '%') or (prior_t, t) == ('=', '('):
+
+            # First check for implicit null values
+            if prior_t in ('=', '%', ','):
+                if t in (',', '/', '&') and not (prior_t, t) == (',', '/'):
+                    append_value(v_values, None, v_idx)
+            else:
+                next_value, t = parse_f90val(tokens, t, prior_t)
+                append_value(v_values, next_value, v_idx)
+
+            # Exit at end of namelist group; otherwise process next value
+            if t in ('/', '&'):
+                break
+            else:
+                t, prior_t = next(tokens), t
+
+    return v_name, v_values, t, prior_t
+
+
+#---
+def append_value(v_values, next_value, v_idx=None):
+    """Update a list of parsed values with a new value."""
+
+    if v_idx:
+        v_i = next(v_idx)
+
+        try:
+            # Default Fortran indexing starts at 1
+            v_values[v_i - 1] = next_value
+        except IndexError:
+            # Expand list to accommodate out-of-range indices
+            size = len(v_values)
+            v_values.extend(None for i in range(size, v_i))
+            v_values[v_i - 1] = next_value
+    else:
+        v_values.append(next_value)
+
+
 #---
 def to_f90str(value):
     """Convert primitive Python types to equivalent Fortran strings"""
@@ -311,7 +360,7 @@ def parse_f90idx(tokens, t, prior_t):
     if t == ':':
         t = next(tokens)
         try:
-            i_end = int(t)
+            i_end = 1 + int(t)
             t = next(tokens)
         except ValueError:
             if t == ':':
@@ -323,7 +372,7 @@ def parse_f90idx(tokens, t, prior_t):
     elif t in idx_end:
         # Replace index with single-index range
         if i_start:
-            i_end = i_start
+            i_end = 1 + i_start
 
     # Stride index
     if t == ':':
@@ -351,7 +400,21 @@ def parse_f90idx(tokens, t, prior_t):
     v_indices.append((idx_triplet))
     t = next(tokens)
 
-    return v_name, v_indices, t
+    return v_indices, t
+
+
+#---
+def merge_values(src, new):
+    """Update a value list with a list of new or updated values."""
+
+    l_min, l_max = (src, new) if len(src) < len(new) else (new, src)
+
+    l_min.extend(None for i in range(len(l_min), len(l_max)))
+
+    for i, val in enumerate(new):
+        new[i] = val if val else src[i]
+
+    return new
 
 
 #---
